function [Res, d] = amp2d(Phi,Psi,y,opts)
%
% Approximate Message Passing (AMP) Algorithm that
%   solves a L1 minimization problem.
%
%     min_x | Psi(x) |1 s.t. |y - Phi(x)|2 < epsilon
%
% Usage:
%       [Res d] = amp2d(Phi,Psi,y,opts)
%
%       Inputs
%           Phi : Measurement Matrix
%           Psi : Sparse Transformation
%                 (Psi=1 when no sparse transformation)
%           y   : Measured Vector
%           opts (a structure containing additional options)
%               opts.x0         - Initial Estimation
%               opts.randshift  - Flag for random shifting
%               opts.tol        - Tolerance Level
%                                 (Optional, default = 10^-2.5)
%               opts.maxIter    - Maximum Iteration (default = 40);
%               opts.verbose    - Number iteration will be displya if 1
%                                 (Optional, default = 0)
%       Outputs
%           Res : Solution x^hat by AMP
%           d   : Sucess Index (1 = Sucess, 0 = Fail)
%
%   Reference:
%           "Message Passing Algorithms for CS"
%           Donoho et al., PNAS 2009;106:18914
%
%
%   Kyung Sung, 4/29/2010
%   ksung@mednet.ucla.edu
%   (c) Stanford University
%

[nx, ny] = size(y);

% opts.x0
if (isfield(opts,'x0') && ~isempty(opts.x0))
    x0 = opts.x0;
else
    x0 = zeros(size(y));
end

% opts.verbose
if (isfield(opts,'verbose') && ~isempty(opts.verbose))
    dispOpt = opts.verbose;
else
    dispOpt = 0;
end

% opts.tol
if (isfield(opts,'tol') && ~isempty(opts.tol))
    tol = opts.tol;
else
    tol = eps;
end

if (isfield(opts,'randshift') && ~isempty(opts.randshift))
    rand_shift = opts.randshift;
else
    rand_shift = 0;
end

if (isfield(opts,'maxIter') && ~isempty(opts.maxIter))
    niter = opts.maxIter;
else
    niter = 40;
end

if (isfield(opts,'sigmaType') && ~isempty(opts.sigmaType))
    sigmaType = opts.sigmaType;
else
    sigmaType = 2;
end

mm = length(find(abs(y)>eps));
rfact = (nx*ny)/mm;

load('OptimumLambdaSigned.mat');  % this file has the optimal values of lambda
delta = 1/rfact;
lambda = interp1(delta_vec,lambda_opt,delta);

% if Psi is numeric, then assume no sparse transformation
no_psi = isa(Psi,'numeric');

% Initial values
wn  = x0;
zn  = y - Phi*x0;
abc = 1;    d = 0;

res_norm = zeros(1,niter+1);
nn       = zeros(1,niter+1);
res_diff = zeros(1,niter);

res_norm(1) = norm(zn(:));
norm_y      = norm(y(:));
nn(1)       = res_norm(1)/norm_y;

while abc < niter+1
    
    % random shift back
    rand_shift_x = randperm(nx,1);
    rand_shift_y = randperm(ny,1);
    
    % First-order Approximate Message Passing
    temp_z = Phi'*zn + wn;

    if (rand_shift == 1)
        temp_z = circshift(temp_z,[rand_shift_x rand_shift_y]);
    end
    
    if (no_psi ~= 1)
        temp_z = Psi*temp_z;
    end
    
    % Compute sigma hat
    if (sigmaType == 1)
        sigma_hat = median(abs(temp_z(:)))/.6745;
    else
        sigma_hat = res_norm(abc)/sqrt(mm);
    end
    
    % If sigma is zero put any VERY small number
    if sigma_hat == 0
        sigma_hat = .1;
    end
    
    % Soft Thresholding
    wn1 = (abs(temp_z)>lambda*sigma_hat).*(abs(temp_z)-lambda*sigma_hat).*sign(temp_z);

    % Compute a sparsity/measurement ratio
    amp_weight = length(find(abs(wn1)>eps))/mm;

    if (no_psi ~= 1)
        wn1 = Psi'*wn1;
    end
    
    % random shift back
    if (rand_shift == 1)
        wn1 = circshift(wn1,[-rand_shift_x -rand_shift_y]);
    end
    
    % Update the residual term
    residual = y - Phi*wn1;
    
    % Normalized data fidelity term
    res_norm(abc+1) = norm(residual(:));
    nn(abc+1)       = res_norm(abc+1)/norm_y;
    res_diff(abc)   = abs(nn(abc+1)-nn(abc));
    
    if (dispOpt)
        fprintf('#iter: %d,   Res. = %.4f,   Res. Diff. = %.5f\n',abc,nn(abc+1),res_diff(abc));
    end
    
    if (abc > 2)
        if (res_diff(abc) < tol)
            d = 1;
            break;
        end
    end
    
    % Update Estimation
    wn = wn1;
    
    if (amp_weight > 1)
        zn = residual + 0.25*zn;
    else
        zn = residual + amp_weight*zn;
    end
    
    abc = abc + 1;
end

if (dispOpt)
    fprintf('------------------------------------------------\n');
end

Res = wn;

