function [Res, d] = lcamp2d(Phi,Psi,y,opts)
%
% Location Constrained Approximate Message Passing (LCAMP) Algorithm that
%   solves a L1 minimization problem.
%
%     min_x | Psi(x) |1 s.t. |y - Phi(x)|2 < epsilon & M Psi(x) = 0
%
%     where M is a sparse location constraint (LC)
%
% Usage:
%       [Res d] = lcamp(Phi,Psi,y,opts)
%
%       Inputs
%           Phi : Measurement Matrix
%           Psi : Sparse Transformation
%                 (Psi=1 when no sparse transformation)
%           y   : Measured Vector
%           opts (a structure containing additional options)
%               opts.x0         - Initial Estimation
%               opts.randshift  - Flag for random shifting
%               opts.tol        - Tolerance Level
%                                 (Optional, default = 10^-2.5)
%               opts.maxIter    - Maximum Iteration (default = 40);
%               opts.verbose    - Number iteration will be displya if 1
%                                 (Optional, default = 0)
%               opts.lc         - Estimated location contraint (LC)
%                                 in image domain
%               opts.lambda     - Flag for soft-thresholing LC
%                                 (Optional, default = 0)
%
%       Outputs
%           Res : Solution x^hat by LCAMP
%           d   : Sucess Index (1 = Sucess, 0 = Fail)
%
%   Reference:
%           "Message Passing Algorithms for CS"
%           Donoho et al., PNAS 2009;106:18914
%
%           "Location constrained approximate message passing for
%           compressed sensing MRI"
%           Sung et al., MRM 2013 
%
%   Kyung Sung, 4/29/2010
%   ksung@mednet.ucla.edu
%   (c) Stanford University
%

[nx, ny] = size(y);

% opts.x0
if (isfield(opts,'x0') && ~isempty(opts.x0))
    x0 = opts.x0;
else
    x0 = zeros(size(y));
end

% opts.verbose
if (isfield(opts,'verbose') && ~isempty(opts.verbose))
    dispOpt = opts.verbose;
else
    dispOpt = 0;
end

% opts.tol
if (isfield(opts,'tol') && ~isempty(opts.tol))
    tol = opts.tol;
else
    tol = eps;
end

if (isfield(opts,'randshift') && ~isempty(opts.randshift))
    rand_shift = opts.randshift;
else
    rand_shift = 0;
end

if (isfield(opts,'maxIter') && ~isempty(opts.maxIter))
    niter = opts.maxIter;
else
    niter = 40;
end

% opts.lc
if (isfield(opts,'lc') && ~isempty(opts.lc))
    locMask = opts.lc;
    locMask = locMask/max(abs(locMask(:)));
else
    locMask = 1;
end

% opts.lambda
if (isfield(opts,'lambda') && ~isempty(opts.lambda))
    lambda_flag = opts.lambda;
else
    lambda_flag = 0;
end

mm = length(find(abs(y)>eps));
rfact = (nx*ny)/mm;

load('OptimumLambdaSigned.mat');  % this file has the optimal values of lambda
delta = 1/rfact;
lambda = interp1(delta_vec,lambda_opt,delta);

% if Psi is numeric, then assume no sparse transformation
no_psi = isa(Psi,'numeric');

% Initial values
wn  = x0;
zn  = y - Phi*x0;
abc = 1;    d = 0;

% Determine an initial thresholding level for LC
% % if (no_psi ~= 1)
% %     tmp_mm = Psi*locMask;
% % else
% %     tmp_mm = locMask;
% % end
% % tmp_mm = tmp_mm/max(abs(tmp_mm(:)));
% % level = 0.1*graythresh(abs(tmp_mm));

res_norm = zeros(1,niter+1);
nn       = zeros(1,niter+1);
res_diff = zeros(1,niter);

res_norm(1) = norm(zn(:));
norm_y      = norm(y(:));
nn(1)       = res_norm(1)/norm_y;

while abc < niter+1
    
    % random shift back
    rand_shift_x = randperm(nx,1);
    rand_shift_y = randperm(ny,1);
    
    % First-order Approximate Message Passing
    temp_z = Phi'*zn + wn;
    
    if (rand_shift == 1)
        temp_z     = circshift(temp_z,[rand_shift_x rand_shift_y]);
        locMask_rs = circshift(locMask,[rand_shift_x rand_shift_y]);
    else
        locMask_rs = locMask;
    end
    
    if (no_psi ~= 1)
        temp_z     = Psi*temp_z;
        locMask_rs = Psi*locMask_rs;
    end
    
    % Compute sigma hat
    sigma_hat = median(abs(temp_z(:)))/.6745;
    
    % If sigma is zero put any VERY small number
    if sigma_hat == 0
        sigma_hat = .1;
    end
    
    % Apply non-zero location mask
    w_sorted = sort(abs(locMask_rs(:)),1,'descend');
    level = abs(w_sorted(round(nx*ny/(0.5*rfact))));
    
    lc_mask = abs(locMask_rs) > level;
    
    if (lambda_flag ~= 0)    % soft-thresholding location mask
        wn1 = (abs(temp_z)>lambda*sigma_hat).*(abs(temp_z)-lambda*sigma_hat).*sign(temp_z);
        wn1(lc_mask == 1) = temp_z(lc_mask == 1);
    else                     % hard-thresholding location mask
        wn1 = temp_z.*lc_mask;
    end
    
    % Compute a sparsity/measurement ratio
    amp_weight = length(find(abs(wn1)>eps))/mm;
    
    if (no_psi ~= 1)
        wn1 = Psi'*wn1;
    end
    
    % random shift back
    if (rand_shift == 1)
        wn1 = circshift(wn1,[-rand_shift_x -rand_shift_y]);
    end
    
    % Update the residual term
    residual = y - Phi*wn1;
    
    % Normalized data fidelity term
    res_norm(abc+1) = norm(residual(:));
    nn(abc+1)       = res_norm(abc+1)/norm_y;
    res_diff(abc)   = abs(nn(abc+1)-nn(abc));
    
    if (dispOpt)
        fprintf('#iter: %d,   Res. = %.4f,   Res. Diff. = %.5f\n',abc,nn(abc+1),res_diff(abc));
    end

    if (abc > 2)
        if (res_diff(abc) < tol)
            d = 1;
            break;
        end
    end
    
    % Update Estimation
    wn = wn1;
    
    if (amp_weight > 1)
        zn = residual + 0.25*zn;
    else
        zn = residual + amp_weight*zn;
    end
    
    abc = abc + 1;
end

if (dispOpt)
    fprintf('------------------------------------------------\n');
end

Res = wn;





